/* wydawca - FTP release synchronization daemon
   Copyright (C) 2007 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "hash.h"

static Hash_table *triplet_table;

static size_t
hash_triplet_hasher (void const *data, unsigned n_buckets)
{
  struct file_triplet const *t = data;
  return hash_string (t->name, n_buckets);
}

/* Compare two strings for equality.  */
static bool
hash_triplet_compare (void const *data1, void const *data2)
{
  struct file_triplet const *t1 = data1;
  struct file_triplet const *t2 = data2;
  return t1->name && t2->name && strcmp (t1->name, t2->name) == 0;
}

/* Reclaim memory storage associated with a table entry */
void
hash_triplet_free (void *data)
{
  int i;
  struct file_triplet *tp = data;
  
  for (i = 0; i < FILE_TYPE_COUNT; i++)
    {
      if (tp->file[i].name)
	free (tp->file[i].name);
    }

  free (tp->directive);
  free (tp->blurb);
  free (tp->tmp);
	  
  free (tp);
}

/* Register a file in the triplet table */
void
register_file (struct file_info *finfo)
{
  struct file_triplet *tp, *ret;
  
  tp = xmalloc (sizeof(*tp) + finfo->root_len + 1);
  memset (tp, 0, sizeof (*tp));
  tp->name = (char*)(tp + 1);
  memcpy (tp->name, finfo->name, finfo->root_len);
  tp->name[finfo->root_len] = 0;
	  
  if (! ((triplet_table
	  || (triplet_table = hash_initialize (0, 0,
					       hash_triplet_hasher,
					       hash_triplet_compare,
					       hash_triplet_free)))
	 && (ret = hash_insert (triplet_table, tp))))
    xalloc_die ();

  if (ret != tp)
    free (tp);
  ret->file[finfo->type] = *finfo;
}

#define SP(s) ((s) ? (s) : "NONE")

/* Return true if any part of the triplet TRP was modified more than
   TTL seconds ago */
static bool
triplet_expired_p (struct file_triplet *trp, time_t ttl)
{
  int i;
  time_t now = time (NULL);

  if (ttl == 0)
    return false;
  
  for (i = 0; i < FILE_TYPE_COUNT; i++)
    {
      if (trp->file[i].name
	  && (now - trp->file[i].mtime) > ttl)
	{
	  if (debug_level)
	    logmsg (LOG_DEBUG, "File %s expired", trp->file[i].name);
	  return true;
	}
    }
  return false;
}

enum triplet_state
  {
    triplet_directive,      /* Short triplet: only a directive is present,
			       but nothing more is required */
    triplet_complete,       /* A complete triplet: all three files are present
			       and have the same owner */
    triplet_incomplete,     /* Incomplete triplet: some files are missing */ 
    triplet_bad,            /* Bad triplet. Should be removed immediately. */
  };


static enum triplet_state
check_triplet_state (struct file_triplet *trp)
{ 
  if (trp->file[file_directive].name)
    {
      if (trp->file[file_dist].name == 0 
	  && trp->file[file_signature].name == 0)
	{
	  if (directive_get_value (trp, "filename", NULL))
	    return triplet_directive ;
	}
      else if (trp->file[file_dist].name
	       && trp->file[file_signature].name)
	{
	  if (trp->file[file_dist].uid == trp->file[file_signature].uid
	      && trp->file[file_dist].uid == trp->file[file_directive].uid)
	    return triplet_complete;
	  else
	    {
	      if (debug_level)
		logmsg (LOG_DEBUG, "%s: invalid triplet: UIDs differ",
			trp->name);
	      return triplet_bad;
	    }
	}
    }
  
  return triplet_incomplete;
}

/* Unlink all parts of the triplet TRP */
static void
remove_triplet (struct file_triplet *trp)
{
  int i;
  
  for (i = 0; i < FILE_TYPE_COUNT; i++)
    {
      if (trp->file[i].name)
	{
	  logmsg (LOG_NOTICE, "Removing %s", trp->file[i].name);
	  if (!dry_run_mode && unlink (trp->file[i].name))
	    logmsg (LOG_ERR, "Cannot remove %s: %s",
		    trp->file[i].name, strerror (errno));
	}
    }
}

/* Process a single triplet from the table */
static bool
triplet_processor (void *data, void *proc_data)
{
  struct file_triplet *trp = data;
  struct directory_pair *dpair = proc_data;

  if (debug_level)
    logmsg (LOG_DEBUG, "FILE %s, DIST=%s, SIG=%s, DIRECTIVE=%s",
	    trp->name,
	    SP (trp->file[file_dist].name),
	    SP (trp->file[file_signature].name),
	    SP (trp->file[file_directive].name));

  if (verify_directive_file (trp, dpair) == 0)
    {
      switch (check_triplet_state (trp))
	{
	case triplet_directive:
	case triplet_complete:
	  if (debug_level)
	    logmsg (LOG_DEBUG, "Processing triplet `%s'", trp->name);
	  process_directives (trp, dpair);
	  return true;

	case triplet_incomplete:
	  if (debug_level)
	    logmsg (LOG_DEBUG, "%s: incomplete triplet", trp->name);
	  /* ignore unless expired (see below); */
	  UPDATE_STATS (STAT_INCOMPLETE_TRIPLETS);
	  break;

	case triplet_bad:
	  UPDATE_STATS (STAT_BAD_TRIPLETS);
	  remove_triplet (trp);
	  return true;
	}
    }
  else
    UPDATE_STATS (STAT_INCOMPLETE_TRIPLETS);
  
  if (triplet_expired_p (trp, dpair->file_sweep_time))
    {
      UPDATE_STATS (STAT_EXPIRED_TRIPLETS);
      remove_triplet (trp);
    }
  
  return true;
}

/* Process all triplets from the table according to the directory pair DPAIR */
void
enumerate_triplets (struct directory_pair *dpair)
{
  if (debug_level)
    logmsg (LOG_DEBUG, "Processing files for %s", dpair->dest_dir);
  if (triplet_table)
    {
      hash_do_for_each (triplet_table, triplet_processor, dpair);
      hash_clear (triplet_table);
    }
}

size_t
count_collected_triplets ()
{
  return triplet_table ? hash_get_n_entries (triplet_table) : 0;
}

