/* wydawca - FTP release synchronization daemon
   Copyright (C) 2007 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3 of the License, or (at your
   option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License along
   with this program. If not, see <http://www.gnu.org/licenses/>. */

#include "wydawca.h"
#include "sql.h"

/* Return the length of the string without trailing whitespace */
size_t
trim_length (const char *str)
{
  size_t len;

  for (len = strlen (str); len > 0 && isspace (str[len-1]); len--)
    ;
  return len;
}

/* Chop off any trailing whitespace. Return the length of the resulting string.
 */
size_t
trim (char *str)
{
  size_t len = trim_length (str);
  str[len] = 0;
  return len;
}

static const char *
find_expansion_char (int c, const struct kw_expansion *exp, size_t nexp)
{
  for (; nexp; nexp--, exp++)
    if (exp->kw[1] == 0 && exp->kw[0] == c)
      return exp->value;
  return NULL;
}

static const char *
find_expansion_word (char *kw, size_t len,
		     const struct kw_expansion *exp, size_t nexp)
{
  for (; nexp; nexp--, exp++)
    if (strlen (exp->kw) == len && memcmp (exp->kw, kw, len) == 0)
      return exp->value;
  return NULL;
}

/* Expand PARAM by replacing %u with the quoted value of USER, %p with that
   of PROJECT and %% with a single %. Return the malloc'ed result.
   FIXME: Should we quote PARAM itself?
*/
char *
expand_param (const char *param, const struct kw_expansion *exp, size_t nexp)
{
  char *p, *q, *res;
  const char *s;
  int len;
  
  if (!param)
    return NULL;
	
  /* Compute resulting param length */
  for (len = 0, p = (char *) param; *p; )
    {
      if (*p == '%')
	{
	  if (p[1] == '%')
	    {
	      len++;
	      p += 2;
	    }
	  else if (p[1] == '{')
	    {
	      char *e = strchr (p + 2, '}');
	      if (e && (s = find_expansion_word (p + 2, e - p - 2, exp, nexp)))
		{
		  len += strlen (s);
		  p = e + 1;
		}
	      else 
		len++, p++;
	    }
	  else if ((s = find_expansion_char (p[1], exp, nexp)) != NULL)
	    {
	      len += strlen (s);
	      p += 2;
	    }
	  else 
	    {
	      /* FIXME: warning? */
	      len++;
	      p++;
	    }
	}
      else
	{
	  len++;
	  p++;
	}
    }
  
  res = xmalloc (len + 1);

  for (p = (char *) param, q = res; *p; )
    {
      if (*p == '%')
	{
	  switch (*++p)
	    {
	    case '%':
	      *q++ = *p++;
	      break;
	      
	    case '{':
	      {
		char *e = strchr (p + 1, '}');
		if (e
		    && (s = find_expansion_word (p + 1, e - p - 1, exp, nexp)))
		  {
		    strcpy (q, s);
		    q += strlen (s);
		    p = e + 1;
		  }
		else 
		  len++, p++;
	      }
	      break;
	      
	    default:
	      if ((s = find_expansion_char (*p, exp, nexp)) != NULL)
		{
		  strcpy (q, s);
		  q += strlen (s);
		  p++;
		}
	      else
		*q++ = *p++;
	    }
	}
      else
	*q++ = *p++;
    }
  *q = 0;
  
  return res;
}

/* Quote non-printable characters in INPUT. Point *OUTPUT to the malloc'ed
   quoted string. Return its length. */
static size_t
quote_string (struct access_method *method, const char *input, char **poutput)
{
  size_t size, len;
  int quote;
  char *output;

  if (!input)
    {
      *poutput = xmalloc (1);
      (*poutput)[0] = 0;
      return 1;
    }
  
  switch (method->type)
    {
    case method_sql:
      len = strlen (input);
      size = 2 * len + 1;
      output = xmalloc (size);
      mysql_real_escape_string (&method->v.sqlconn->mysql, output, input, len);
      size = strlen (output);
      break;

    default:
      size = argcv_quoted_length (input, &quote);
      output = xmalloc (size);
      argcv_quote_copy (output, input);
      break;
    }

  *poutput = output;
  return size;
}

static void
escape_kwexp (struct access_method *method, struct kw_expansion *exp,
	      size_t nexp)
{
  size_t i;

  for (i = 0; i < nexp; i++)
    quote_string (method, exp[i].value, &exp[i].value);
}

void
free_kwexp (struct kw_expansion *exp, size_t nexp)
{
  size_t i;

  for (i = 0; i < nexp; i++)
    free (exp[i].value);
}



/* Verify if USER has upload rights on the directory (project) requested
   by TRP */
int
check_access_rights (struct file_triplet *trp, struct directory_pair *dpair,
		     const char *user)
{
  struct access_method *method = dpair->verify_method;
  char *project;
  const char *directory;
  char *p;
  int rc;
  char *command;
  const char *result;
  struct kw_expansion kwexp[2];
  
  if (directive_get_value (trp, "directory", &directory))
    {
      logmsg (LOG_ERR, "%s: missing `directory' directive",
	      trp->file[file_directive].name);
      return 1;
    }
  p = strchr (directory, '/');
  if (p)
    {
      size_t len = p - directory;
      project = xmalloc (len + 1);
      memcpy (project, directory, len);
      project[len] = 0;
    }
  else
    project = xstrdup (directory);

  if (debug_level)
    logmsg (LOG_DEBUG, "verifying access rights for user %s to project %s",
	    user, project);

  kwexp[0].kw = "u";
  kwexp[0].value = (char*) user;
  kwexp[1].kw = "p";
  kwexp[1].value = project;

  escape_kwexp (method, kwexp, 2);
  command = expand_param (method->param[1], kwexp, 2);
  free_kwexp (kwexp, 2);
  
  rc = method_run (method, command);
  free (command);
  if (rc)
    {
      logmsg (LOG_ERR, "Cannot verify access rights for %s on %s",
	      user, project);
      free (project);
      return 1;
    }

  result = method_result (method);
  if (strcmp (result, user))
    {
      /* FIXME: Notify */
      UPDATE_STATS (STAT_ACCESS_VIOLATIONS);
      logmsg (LOG_ERR, "%s: %s has no rights on %s",
	      trp->name, user, project);
      free (project);
      return 1;
    }
  free (project);
	      
  return 0;
}

int
verify_directive_file (struct file_triplet *trp, struct directory_pair *dpair)
{
  struct passwd *pw;
  char *user_name;
  char *command;
  struct access_method *method = dpair->gpg_key_method;
  const char *pubkey;
  int rc;
  struct kw_expansion kwexp[1];
  
  if (!trp->file[file_directive].name)
    return 1;
  
  pw = getpwuid (TRIPLET_UID (trp));
  if (!pw)
    {
      logmsg (LOG_ERR, "%s: getpwuid failed: %s",
	      trp->name, strerror (errno));
      return 1;
    }
  trp->gid = pw->pw_gid;
  user_name = pw->pw_name;

  kwexp[0].kw = "u";
  kwexp[0].value = user_name;
  escape_kwexp (method, kwexp, 1);
  command = expand_param (method->param[1], kwexp, 1);
  free_kwexp (kwexp, 1);
  
  rc = method_run (method, command); 
  free (command);
  if (rc)
    {
      logmsg (LOG_ERR, "Cannot get PGP key for %s", user_name);
      return 1;
    }
  
  pubkey = method_result (method);
  if (pubkey[0] == 0)
    {
      logmsg (LOG_ERR, "No GPG key for %s", user_name);
      return 1;
    }

  if (debug_level > 3)
    logmsg (LOG_DEBUG, "Public key: %s", pubkey);

  if (verify_directive_signature (trp, dpair, pubkey))
    {
      logmsg (LOG_ERR, "invalid signature for %s", trp->name);
      return 1;
    }
  else if (debug_level)
    logmsg (LOG_DEBUG, "%s: directive file signature OK", trp->name);

  if (debug_level > 1)
    {
      int i;
      for (i = 0; trp->directive[i]; i++)
	logmsg (LOG_DEBUG, "directive[%d] = %s", i, trp->directive[i]);
    }

  if (verify_directive_format (trp))
    return 1;

  if (check_access_rights (trp, dpair, user_name))
    /* FIXME: Notify user / admin */
    return 1;

  return 0;
}

